#include <assert.h>
#include <stdio.h>
#include <stdint.h>
#include <omp.h>
#include "prime.h"
#include "bitmap.h"
#include "ompadm.h"

/*
	Copyright (c) 2014 Andrew V. Sutherland
	See LICENSE file for license details
*/

#define B		7		// with B=7 and L=6 we use 1KB bitmasks, which seems to be otpimal for dense admissibility testing
#define L		6		// use 2^L bit words in bitmask array of 2^B words

// determines whether the sequence of primes p_{m+1}, ..., p_{m+k0} is admissible at p_{m+1}
int zhang_admissible_mod_firstp (prime_table_t tab, long k0, long m)
{
	uint64_t b[1<<B];
	unsigned char *gap, *egap;
	register long j, r, x, firstp, lastp;
	register int i;
	
	firstp = prime_table_nth_prime (tab, m+1);
	lastp = prime_table_nth_prime (tab, m+k0);
	assert ( firstp && lastp );
	egap = tab->gaps + m + k0 - tab->si;
	for ( j = 0 ;  j < (firstp>>(L+B)) ; j++ ) {
		memset (b, 0, (1<<B)*sizeof(*b));
		for ( x = firstp+(j<<(L+B)) ; x <= lastp ; x+= firstp ) {
			for ( r = prime_table_next_gap (tab, &gap, x)-x ; r < (1L<<(L+B)) && gap <= egap ; r+=2*(*gap++) )
				b[(r>>L)&((1<<B)-1)] |= (1UL << (r&((1<<L)-1)));
			for ( i = 0; i < (1<<B) ; i++ ) if ( ~b[i] ) break;
			if ( i == (1<<B) ) break;
		}
		if ( x > lastp ) return 1;
	}
	for ( j <<= L+B ; j < firstp ; j++ ) {
		for ( x = firstp+j ; x <= lastp ; x+= firstp ) if ( prime_table_next_gap (tab, &gap, x) == x ) break;
		if ( x > lastp ) return 1;
	}
	return 0;
}

// Heuristically checks whether p_{n+1},...,p_{n+k0} is admissible for all n >= m by verifying for n in [m,m+checks) that the sequence p_{n+1},...,p_{n+k0} is admissible modulo p_{n+1}
int zhang_probe (prime_table_t tab, long k0, long m, long checks, int threads)
{
	int quit = 0;
	
	#pragma omp parallel num_threads(threads)
	{
		int tid = omp_get_thread_num ();
		int sts = 1;
		register long endm, n;
		
		endm = m+checks;
		for ( n = m+tid ; sts && n < endm ; n += threads ) {
			sts = zhang_admissible_mod_firstp (tab, k0, n);
			#pragma omp critical(probe)
			{ if ( ! sts ) quit = 1; if ( quit ) sts = 0; }
		}
	}
	return !quit;
}

// Heuristically determines the least m for which p_{n+1},...,p_{n+k0} is admissible for all n >= m
long zhang_best_m (prime_table_t tab, long k0, double begin, int threads)
{
	long m, m0, m1, jump, checks, maxbad;

	printf ("Heuristic search for minimal Zhang m for k0=%ld\n", k0);
	// checks ~ sqrt(k0) is a heuristic estimate for the number of admissibility checks needed to ensure admissibility (being admissible mod p_{m+1},...,p_{m+checks} usually means being admissible modulo all primes)
	// Note that we need to verify our final result in any case, the main loop below does not produce a provable correct result in any case.
	m = prime_table_pi(tab, k0);
	checks = threads *(sqrt(k0)/threads+1);
	jump = 16;
	m0 = 1;								// known lower bound on m
	m1 = 2*m/log(m);					 	// heuristic guess of an approximate upper bound on the minimal m, will be increased if necessary
	while ( m1+checks < m && ! zhang_probe  (tab, k0, m1, checks, threads) ) m1 += jump*checks;
	if ( m1+checks < m ) {
		printf ("  m <= %ld (%.1fs)\n", m1, omp_get_wtime()-begin);
		for ( jump /= 2 ; jump > 1 ; jump /= 2 ) {
			for ( m = m1-jump*checks ; m > m0 && zhang_probe  (tab, k0, m, checks, threads) ; m -= jump*checks ) printf ("  m <= %ld (%.1fs)\n", m1=m, omp_get_wtime()-begin);
			if ( m > m0 ) printf ("  m >  %ld (%.1fs)\n", m0=m, omp_get_wtime()-begin);
		}
		m=m1;
	}
	maxbad = 0;

	#pragma omp parallel num_threads(threads)
	{
		double start;
		int sts, quit = 0;
		long n;
		
		start = omp_get_wtime();
		for ( n = m-1-omp_get_thread_num() ; !quit && n > 0 ; n -= threads ) {
			sts = zhang_admissible_mod_firstp (tab, k0, n);
			#pragma omp critical(mainloop)
			{
				if ( ! sts && n > maxbad ) maxbad = n;
				if ( n <= maxbad ) quit = 1;
			}
			#pragma omp master
			{
				double end = omp_get_wtime();
				if ( end-start > 3 ) { printf ("  m ~ %ld (%.1fs)\n", n, end-begin); start = end; }
			}
		}
	}
	m = maxbad+1;
	printf ("Apparently minimal m = %ld yields diameter %ld (%.1fs)\n", m, prime_table_nth_prime(tab, m+k0)-prime_table_nth_prime(tab, m+1), omp_get_wtime()-begin);
	return m;
}

// determines whether the sequence -p_{m+k0/2+1-i},-p_{m+1},-1,1,p_{m+1}, ..., p_{m+(k0+1)/2-1+i} is admissible modulo p_{m+1}
// pB points to an integer between 0 and MAXB that is used to optimize the test as the density mod p changes
// requires tab to start at p_{m+1} (this is verified)
int ahr_admissible_mod_pi (prime_table_t tab, long k0, long m, long i, long pi)
{
	uint64_t b[1<<B];
	unsigned char *gap;
	register unsigned char *sgap, *egapu, *egapd;
	register long j, r, x, lastpu, lastpd, p;
	register int n;
	
	p = prime_table_nth_prime (tab, pi);
	lastpu = prime_table_nth_prime (tab, m+(k0+1)/2-1+i);
	lastpd = prime_table_nth_prime (tab, m+k0/2-1-i);
	sgap = tab->gaps + m + 1 - tab->si;
	egapu = tab->gaps + m + (k0+1)/2-1+i - tab->si;
	egapd = tab->gaps + m + k0/2-1-i - tab->si;

	// Note that we do not need to consider -1 in the main loop, since it only checks residues 0,...,m-1 mod p, where m is the largest multiple of 2^(L+B) less than p, so we always have m-1 < p-1 (since p is odd and m is even)
	for ( j = 0 ;  j < (p>>(L+B)) ; j++ ) {
		memset (b, 0, (1<<B)*sizeof(*b));
		if ( ! j ) b[0] = 1UL<<1;											// we know 1 is in the sequence
		for ( x = (j<<(L+B)) ; x <= lastpu ; x += p ) {
			for ( r = prime_table_next_gap (tab, &gap, x)-x ; r < (1L<<(L+B)) && gap <= egapu ; r+=2*(*gap++) )
				if ( gap >= sgap ) b[(r>>L)&((1<<B)-1)] |= (1UL << (r&((1<<L)-1)));
			for ( n = 0; n < (1<<B) ; n++ ) if ( ~b[n] ) break;
			if ( n == (1<<B) ) break;
		}
		if ( x <= lastpu ) continue;
		for ( x = p-((j+1)<<(L+B))+1 ; x <= lastpd ; x += p ) {
			for ( r = (1L<<(L+B))-1 + x-prime_table_next_gap (tab, &gap, x) ; r >= 0 && gap <= egapd ; r-=2*(*gap++) )
				if ( gap >= sgap ) b[(r>>L)&((1<<B)-1)] |= (1UL << (r&((1<<L)-1)));
			for ( n = 0; n < (1<<B) ; n++ ) if ( ~b[n] ) break;
			if ( n == (1<<B) ) break;
		}
		if ( x > lastpd ) return 1;
	}
	for ( j <<= L+B ; j < p-1 ; j++ ) {	// we know -1 is in the sequence
		if ( j == 1 ) continue;		// we know 1 is in the squence
		for ( x = j ; x <= lastpu ; x += p ) if ( prime_table_next_gap (tab, &gap, x) == x && gap >= sgap ) break;
		if ( x > lastpu ) {
			for ( x = p-j ; x <= lastpd ; x += p ) if ( prime_table_next_gap (tab, &gap, x) == x && gap >= sgap ) break;
			if ( x > lastpd ) return 1;
		}
	}
	return 0;
}

int hr_probe (prime_table_t tab, long k0, long m, long checks, int threads)
{
	int quit = 0;
	
	#pragma omp parallel num_threads(threads)
	{
		int tid = omp_get_thread_num ();
		int sts = 1;
		long endm, n;
		
		endm = m+checks;
		for ( n = m+tid ; sts && n < endm ; n += threads ) {
			sts = hr_admissible_mod_firstp (tab, k0, n);
			#pragma omp critical(probe)
			{ if ( ! sts ) quit = 1; if ( quit ) sts = 0; }
		}
	}
	return !quit;
}

// determines a heuristically optimal m that minimizes the diameter of an admissible HR-sequence for a given k0 (for small k0 it will not necessarily be optimal, but it does guarantee that m-1 does not work)
long hr_best_m (prime_table_t tab, long k0, double begin, int threads)
{
	int jump;
	long maxbad;
	long m, m0, m1, checks;
	
	printf ("Heuristic search for minimal Hensley-Richards m for k0=%ld\n", k0);

	// checks ~ sqrt(k0) is a heuristic estimate for the number of admissibility checks needed to ensure admissibility (being admissible mod p_{m+1},...,p_{m+checks} usually means being admissible modulo all primes)
	// Note that we need to verify our final result in any case, the main loop below does not produce a provable correct result in any case.
	checks = threads *(sqrt(k0)/threads+1);
	jump = 16;
	m = prime_table_pi (tab, k0);
	m0 = 1;								// known lower bound on m
	m1=2*m/log(m);					 	// heuristic guess of an approximate upper bound on the minimal m, will be increased if necessary
	while ( m1+checks < m && ! hr_probe  (tab, k0, m1, checks, threads) ) m1 += jump*checks;
	if ( m1+checks < m ) {
		printf ("  m <= %ld (%.1fs)\n", m1, omp_get_wtime()-begin);
		for ( jump /= 2 ; jump > 1 ; jump /= 2 ) {
			for ( m = m1-jump*checks ; m > m0 && hr_probe  (tab, k0, m, checks, threads) ; m -= jump*checks ) printf ("  m <= %ld (%.1fs)\n", m1=m, omp_get_wtime()-begin);
			if ( m > m0 ) printf ("  m >  %ld (%.1fs)\n", m0=m, omp_get_wtime()-begin);
		}
		m=m1;
	}
	maxbad = 0;
	#pragma omp parallel num_threads(threads)
	{
		double start;
		int sts, quit = 0;
		long n;
		
		start = omp_get_wtime();
		for ( n = m-1-omp_get_thread_num() ; !quit && n > 0 ; n -= threads ) {
			sts = hr_admissible_mod_firstp (tab, k0, n);
			#pragma omp critical(mainloop)
			{
				if ( ! sts && n > maxbad ) maxbad = n;
				if ( n <= maxbad ) quit = 1;
			}
			#pragma omp master
			{
				double end = omp_get_wtime();
				if ( end-start > 3 ) { printf ("  m ~ %ld (%.1fs)\n", n, end-begin); start = end; }
			}
		}
	}
	m = maxbad+1;
	printf ("Apparently minimal m = %ld (%.1fs)\n", m, omp_get_wtime()-begin);
	return m;
}

int gap_table_admissible_mod_p (gap_table_t tab, long p)
{
	uint64_t b[(1<<B)+1];
	unsigned char *gap;
	register unsigned char *egap;
	register long j, r, x, o;
	register int n;

	assert (p);
	egap = tab->gaps + tab->ti - tab->si;
	b[1<<B] = 0;
	o = tab->sp - mod(tab->sp, p) - p;		// o is the greatest multiple of p such that o+p <= sp
	for ( j = 0 ;  j < (p>>(L+B)) ; j++ ) {
		memset (b, 0, (1<<B)*sizeof(*b));
		n = 0;
		for ( x = o + (j<<(L+B)) ; x <= tab->tp ; x += p ) {
			for ( r = gap_table_next_gap (tab, &gap, x)-x ; r < (1L<<(L+B)) && gap <= egap ; r+=2*(*gap++) )
				b[(r>>L)&((1<<B)-1)] |= (1UL << (r&((1<<L)-1)));
			while ( b[n] == 0xFFFFFFFFFFFFFFFFUL ) n++;
			if ( n == (1<<B) ) break;
		}
		if ( x > tab->tp ) return 1;
	}
	for ( j <<= L+B ; j < p ; j++ ) {
		for ( x = o + j ; x <= tab->tp ; x += p ) if ( gap_table_next_gap (tab, &gap, x) == x ) break;
		if ( x > tab->tp ) return 1;
	}
	return 0;
}

// return the least minimally occupied residue modulo p, or -1 if sequence is admissible mod p
long gap_table_best_residue (gap_table_t tab, long p)
{
	register int *map, *ap, *ep;
	register unsigned char *gap, *egap;
	register long a;
	register int n;

	assert ( p > 510 );
	egap = tab->gaps + tab->ti - tab->si;
	map = calloc (p, sizeof(*map));  ep = map+p;
	for ( ap = map + mod (tab->sp, p), gap = tab->gaps ; gap <= egap ; ap += 2*(*gap++) ) { if ( ap >= ep ) ap -= p ; (*ap)++; }
	for ( n = map[0], a = 0, ap = map+1 ; ap < ep ; ap++ ) if ( *ap < n ) { n = *ap; a = ap-map; }
	free (map);
	return ( n ? a : -1 );
}

// checks admissibility mod p_n for n=m+1 to m+checks, returns 0 if admissible at all p_n, ow least n for which this is not true
long gap_table_check_admissibility (gap_table_t gtab, prime_table_t ptab, long m, long checks, long maxm, int threads)
{
	long minbadm = 0;
	
	assert ( m <= maxm );
	if ( checks > maxm-m ) checks = maxm-m;
	#pragma omp parallel num_threads(threads)
	{
		int tid = omp_get_thread_num ();
		int sts = 1;
		long n;
		
		for ( n = m+1+tid ; sts && n <= m+checks ; n += threads ) {
			sts = gap_table_admissible_mod_p (gtab, prime_table_nth_prime (ptab, n));
			#pragma omp critical(gap_table_check_admissibility)
			{ if ( ! sts && (!minbadm || n < minbadm) ) minbadm = n; if ( minbadm && minbadm <= n ) sts = 0; }
		}
	}
	return minbadm;
}

// constructs a bitmap whose nth entry is 1 if the sequence defined by gtab is admissible modulo p_{m+1+n} for n from 0 to checks-1 (for m+1+n > maxm admissibility bit is set)
long gap_table_admissibility_map (uint64_t *bm, gap_table_t gtab, prime_table_t ptab, long m, long checks, long maxm, int threads)
{	
	long w = 0;
	
	assert ( m <= maxm );
	if ( checks > maxm-m ) { bm_setall (bm, checks);  checks = maxm-m; }
	bm_clear (bm, checks);
	#pragma omp parallel num_threads(threads)
	{
		int tid = omp_get_thread_num ();
		int sts;
		long n;
		
		for ( n = tid ; n < checks ; n += threads ) {
			sts = gap_table_admissible_mod_p (gtab, prime_table_nth_prime (ptab, m+1+n));
			if ( sts ) {
				#pragma omp critical(gap_table_admissibility_map)
				{ bm_set (bm, n); w++; }
			}
		}
	}
	return w;
}

// sets rm[n] to the least occupied residue class mod p_{m+1+n}, or to -1 if admissible at p_{m+1+n}, for n from 0 to checks-1
long gap_table_residue_map (long *rm, gap_table_t gtab, prime_table_t ptab, long m, long checks, long maxm, int threads)
{
	long w = 0;
	
	assert ( m <= maxm );
	if ( checks > maxm-m ) { for ( long i = maxm-m ; i < checks ; i++ ) { rm[i] = -1; w++; }  checks = maxm-m; }
	#pragma omp parallel num_threads(threads)
	{
		int tid = omp_get_thread_num ();
		long r, n;
		
		for ( n = tid ; n < checks ; n += threads ) {
			r = gap_table_best_residue (gtab, prime_table_nth_prime (ptab, m+1+n));
			rm[n] = r;
		}
	}
	for ( long i = 0 ; i < checks ; i++ ) if ( rm[i] == -1 ) w++;
	return w;
}

long gap_table_best_start (gap_table_t gtab, long k0)
{
	register long i, delta, mini, mindelta;
	
	mindelta = delta = mini = 0;
	for ( i=0 ; i+k0 <= gtab->ti-gtab->si ; i++ ) { delta += gtab->gaps[i+k0-1]-gtab->gaps[i]; if ( delta < mindelta ) { mindelta = delta; mini = i+1; } }
	return mini;
}


// sieve bitmap at 1 mod 2 and 0 mod p_n for n in [2,m]
static inline void bitmap_schinzel_sieve (uint64_t *b, long blen, long s, long m, prime_table_t ptab)
{
	register long i, j, p;
	for ( i = mod(1-s,2) ; i < blen ; i+= 2 ) bm_set (b, i);
	for ( j = 2 ; j <= m ; j++ ) {
		p = prime_table_nth_prime (ptab, j);
		for ( i = mod(-s,p) ; i < blen ; i+= p ) bm_set (b, i);
	}
}

// sieve bitmap 0 mod p_n for n in [m0+1,m1], and if nlist is non-null, at r[i] mod p_i for i in nlist (which must be null-terminated)
// if rlist is null, then r[i] = 0 for all i
static inline void bitmap_extended_sieve (uint64_t *b, long blen, long s, long m0, long m1, long *nlist, long *rlist, prime_table_t ptab)
{
	register long i, j, p;

	for ( j = m0+1 ; j <= m1 ; j++ ) {
		p = prime_table_nth_prime (ptab, j);
		for ( i = mod(-s,p) ; i < blen ; i+= p ) bm_set (b, i);
	}
	if ( ! nlist ) return;
	if ( ! rlist ) {
		while ( *nlist ) {
			p = prime_table_nth_prime (ptab, *nlist++);
			for ( i = mod (-s, p) ; i < blen ; i += p ) bm_set (b, i);
		}
		return;
	}
	while ( *nlist ) {
		p = prime_table_nth_prime (ptab, *nlist++);
		assert (p);
		for ( i = mod (*rlist++ -s, p) ; i < blen ; i += p ) bm_set (b, i);
	}
}

// sieves the interval [s,s+d-1] of 1 mod 2 and 0 mod p for primes 3,...,p_n, and optionally at rlist[i] mod p_{nlist[i]}, where nlist is a zero-terminated list of prime indices n > m
// if rlist is null, sieves at 0 mod each p_{nnlist[i]}
// survivors are encoded as a gap sequence in which each byte holds gaplen/2 for gaps of length < 510,
// while gaps of length 510*a+b > 510 (with b in [2,510] nonzero) are encoded as a zeros followed by b/2, *pglistlen is the number of bytes in the gap sequence (which may be larger than the number of gaps)
// *first is set to the value of the first survivor, which may be greater than s, *last is set to the value of the last survivor, which may be less than s+d-1
// returns the number of survivors (equal to *pglistlen+1 in the case that all gaps are < 510)
// pglist is guaranteed to have at least one extra bytes allocated at the end (used for null terminator in the case that all gaps are < 510, needed to create a gap table)
long build_gap_list (long *first, long *last, unsigned char **pglist, long *pglistlen, long s, long d, long m, long*nlist, long *rlist, prime_table_t ptab, int threads)
{
	uint64_t *b0;
	struct { unsigned char *gaps; long k, len, first, last; } *gaplists;
	register unsigned char *gaps, *egap;
	register long blen, bcnt, i, j, k;

	assert ( m >= 6 );
	j = m;  if ( nlist ) for ( i = 0 ; nlist[i] ; i++ ) if ( nlist[i] > j ) j = nlist[i];
	blen = 2*prime_table_nth_prime (ptab, j);			// aim for block size that is big enough so that we are sieving at least 2 entries for each prime
	assert (blen);									// if j is out of range of prime table we will have problems
	if ( d/blen < threads ) blen = d/threads;				// but try to make sure we can give work to every thread
	blen += 64L*15015 - (blen%(64L*15015));			// round b up to a multiple of 64*3*5*7*11*13
	bcnt = d/blen + ( (d%blen) ? 1 : 0 );

	b0 = bm_alloc (blen);
	bitmap_schinzel_sieve (b0, blen, s, 6, ptab);

	gaplists = calloc (bcnt, sizeof(*gaplists));
	#pragma omp parallel num_threads(threads)
	{
		uint64_t *b;
		register unsigned char *g, *h;
		int tid = omp_get_thread_num ();
		register long j, k, p, n, e;
		long i;
		
		b = bm_alloc (blen);
		for ( i = tid ; i < bcnt ; i += threads ) {
			memcpy (b, b0, blen>>3);
			e = i < bcnt-1 ? blen : d-i*blen;
			bitmap_extended_sieve (b, e, s+i*blen, 6, m, nlist, rlist, ptab);
			p = gaplists[i].first = bm_min_unset (b, e);
			if ( p == e ) { gaplists[i].k = 0;  gaplists[i].first = e;  gaplists[i].last = gaplists[i].len = 0; break; }
			assert ( p < e );
			j = e-bm_weight (b, e) + blen/510;
			h = g = malloc (j);
			for ( k = 1 ; ; p=n, k++ ) {
				n = bm_next_unset (b, p+1, e);
				if ( n == e ) break;
				assert ( n < e );
				for ( j = n-p ; j > 510 ; j -= 510 )  *h++ = 0;
				assert (!(j&1));
				*h++ = j>>1;
			}
			gaplists[i].last = n-p;
			gaplists[i].k = k;
			gaplists[i].len = h-g;
			gaplists[i].gaps = g;
		}
		bm_free (b);
	}
	bm_free (b0);
	for ( i = j = k = 0 ; i < bcnt-1 ; i++ ) {
		if ( ! gaplists[i].k ) { gaplists[i+1].first = gaplists[i].first; continue; }			// if no survivors, push it into the next block and skip ahead
		gaplists[i+1].first += gaplists[i].last;									// push last gap of current block to first gap of next block
		if ( i ) j += gaplists[i].first/510+1;
		j += gaplists[i].len;  k += gaplists[i].k;
	}
	if ( gaplists[i].k ) {
		if ( i ) j += gaplists[i].first/510+1;
		j += gaplists[i].len;  k += gaplists[i].k;		
	} else {
		gaplists[i].last = gaplists[i].first;  gaplists[i].first = 0;						// if last block has no survivors, push everything into last
	}
	assert (k);
	gaps = malloc (j+1);													// add a byte to include room for null-terminator needed for gap tables
	
	// skip ahead to first survivor, then copy first list of gaps from there
	for ( i = 0 ; i < bcnt && ! gaplists[0].k ; i++ );
	*first = s+gaplists[i].first;
	*last = s+d-gaplists[bcnt-1].last;
	memcpy (gaps, gaplists[i].gaps, gaplists[i].len);  egap = gaps + gaplists[i].len;
	free (gaplists[i].gaps);
	for ( i++ ; i < bcnt ; i++ ) {
		if ( ! gaplists[i].k ) continue;
		for ( j = gaplists[i].first ; j > 510 ; j -= 510 ) *egap++ = 0;
		assert (!(j&1));
		*egap++ = j>>1;
		memcpy (egap, gaplists[i].gaps, gaplists[i].len);
		egap += gaplists[i].len;
		free (gaplists[i].gaps);
	}
	*pglist = gaps;
	*pglistlen = egap-gaps;
	free (gaplists);
	return k;
}


// sieves the interval [s,s+d-1] of 1 mod 2 and 0 mod p for primes 3,...,p_n, and optionally at rlist[i] mod p_{nlist[i]}, where nlist is a list of prime indicies > n
// survivors are encoded as a gap sequence (each byte holds gap/2, requires maxgap < 510, this is checked)
long build_gap_table (gap_table_t gtab, long s, long d, long m, long*nlist, long *rlist, prime_table_t ptab, int threads)
{
	unsigned char *glist;
	long k, first, last, glistlen;
	
	k = build_gap_list (&first, &last, &glist, &glistlen, s, d, m, nlist, rlist, ptab, threads);
	assert ( glistlen+1 == k );			// check that all gaps are encoded in a single byte
	glist[glistlen] = 0;					// add terminating null, glist is guaranteed to have space for it
	gap_table_init (gtab, first, glist, glistlen);
	return k;
}

// increases d as required to ensure at least k0 survivors, decreases d if it looks too big
// returns number of survivors
long build_gap_table_k0 (gap_table_t gtab, long k0, long s, long *pd, long m,  long *nlist, long *rlist, prime_table_t ptab, double begin, int threads)
{
	long n, newd, len, max, d;
	
	d = *pd;
	len = max = 0;
	if ( nlist ) for ( ; nlist[len] ; len++ ) if ( nlist[len] > max ) max = nlist[len];
	for (;;) {
		n = build_gap_table (gtab, s, *pd, m, nlist, rlist, ptab, threads);
		if ( n >= k0 ) {
			// trim d if it looks too large
			while ( (double) n/k0 > 1.001 ) {
				long newd = ((((long)(0.999*(*pd)+1))>>6)+1)<<6;
				if ( newd == *pd ) break;
  				gap_table_clear (gtab);
				n = build_gap_table (gtab, s, newd, m,  nlist, rlist, ptab, threads);
				if ( n < k0 ) {
					gap_table_clear (gtab);
					n = build_gap_table (gtab, s, *pd, m,  nlist, rlist, ptab, threads);
					break;
				}
				*pd = newd;
			}
			if ( *pd != d ) printf ("%s d to %ld (%.1fs)\n", (d < *pd ? "increased" : "decreased" ), *pd, omp_get_wtime()-begin);
			return n;
		}
		gap_table_clear (gtab);
		newd = ((((long)(1.001*(*pd)+1))>>6)+1)<<6;
		if ( newd == *pd ) newd += 64;
		*pd = newd;
	}
}

// finds the least m such that sieving [s,s+d] at 1 mod 2 and 0 mod p_n for n <= m yields an admissible tuple
// adjusts d as required to ensure # survivors is slightly greater then or equal to k0)
long schinzel_best_m (prime_table_t ptab, long k0, long s, long *pd, double begin, int threads)
{
	gap_table_t gtab;
	long m, badm, maxm, checks;
	
	printf ("Heuristic search for minimal Schinzel m for k0=%ld using shift s=%ld\n", k0, s);
	
	// start by getting a reasonably close estimate of m, which we empirically expect to be around 2*k0/log(k0)^2
	checks = threads *(sqrt(k0)/threads+1);
	m = (long) (2*k0/(log(k0)*log(k0)));
	if ( m < 6 ) m = 6;
	maxm = prime_table_pi (ptab, k0);
	assert (maxm);
	schinzel_gap_table_k0 (gtab, k0, s, pd, m, ptab, begin, threads);
	badm = gap_table_check_admissibility (gtab, ptab, m, checks, maxm, threads);
	gap_table_clear (gtab);
	if ( badm ) {
		long lastbadm = badm;
		printf ("  m > %ld (%.1fs)\n", m, omp_get_wtime()-begin);
		for ( m += checks ;  m < maxm && badm ; m += checks ) {
			lastbadm = badm;
			schinzel_gap_table_k0 (gtab, k0, s, pd, m, ptab, begin, threads);
			badm = gap_table_check_admissibility (gtab, ptab, m, checks, maxm, threads);
			gap_table_clear (gtab);
			if ( badm ) printf ("  m > %ld (%.1fs)\n", m, omp_get_wtime()-begin);
		}
		m += 2*checks;
		if ( m > maxm ) m = maxm;
		badm = lastbadm;
	} else {
		printf ("  m <= %ld (%.1fs)\n", m, omp_get_wtime()-begin);
		for ( m -= checks ; m >= 6 && ! badm ; m -= checks ) {
			schinzel_gap_table_k0 (gtab, k0, s, pd, m, ptab, begin, threads);
			badm = gap_table_check_admissibility (gtab, ptab, m, checks, maxm, threads);
			gap_table_clear (gtab);
			if ( ! badm ) printf ("  m <= %ld (%.1fs)\n", m, omp_get_wtime()-begin);
		}
		if ( m < 6 ) m = 6;
	}
	
	// Now work our way up to find the least m that appears to work
	while ( badm ) {
		m = badm;
		schinzel_gap_table_k0 (gtab, k0, s, pd, m, ptab, begin, threads);
		badm = gap_table_check_admissibility (gtab, ptab, m, 2*checks, maxm, threads);
		gap_table_clear (gtab);		
		if ( badm ) printf ("  m > %ld (%.1fs)\n", m, omp_get_wtime()-begin);
	}
	// just in case we overshot
	while ( ! badm ) {
		m--;
		if ( m <= 6 ) { m = 6; break; }
		schinzel_gap_table_k0 (gtab, k0, s, pd, m, ptab, begin, threads);
		badm = gap_table_check_admissibility (gtab, ptab, m, 2*checks, maxm, threads);
		gap_table_clear (gtab);
		if ( badm ) m++;
	}
	printf ("Apparently miminal m = %ld (%.1fs)\n", m, omp_get_wtime()-begin);

	return m;
}

long build_gap_sequence (gap_sequence_t seq, long s, long d, long m,  long *nlist, long *rlist, prime_table_t ptab, int threads)
{
	long len;
	register unsigned char *gap;
	
	seq->k = build_gap_list (&seq->s, &seq->t, &seq->gaps, &len, s, d, m, nlist, rlist, ptab, threads);
	seq->egap = seq->gaps + len;
	if ( len != seq->k-1 ) {
		assert ( len > seq->k-1 );
		assert (*(seq->egap-1));
		seq->firstz = seq->gaps + strnlen ((char*)seq->gaps, len);
		assert (seq->firstz < seq->egap);
		for ( gap = seq->egap-2 ; *gap ; gap-- );
		seq->endz = gap+2;
	} else {
		seq->firstz = seq->endz = seq->egap;
	}
	return seq->k;
}

void clear_gap_sequence (gap_sequence_t seq)
	{  free (seq->gaps);  seq->gaps = 0;  }

	
// increases d as required to ensure at least k0 survivors, decreases d if it looks too big
// returns number of survivors
long build_gap_sequence_k0 (gap_sequence_t seq, long k0, long s, long *pd, long m,  long *nlist, long *rlist, prime_table_t ptab, double begin, int threads)
{
	long n, newd, len, max, d, oldn;
	
	d = *pd;
	len = max = oldn = 0;
	if ( nlist ) for ( ; nlist[len] ; len++ ) if ( nlist[len] > max ) max = nlist[len];
	for (;;) {
		n = build_gap_sequence (seq, s, *pd, m, nlist, rlist, ptab, threads);
		if ( oldn && n <= oldn ) printf ("Warning, increased d to %ld with no change in increase in survivors %ld\n", *pd, n);
		oldn = n;
		if ( n >= k0 ) {
			// trim d if it looks too large
			while ( (double) n/k0 > 1.001 ) {
				long newd = ((((long)(0.999*(*pd)+1))>>6)+1)<<6;
				if ( newd == *pd ) break;
  				clear_gap_sequence (seq);
				n = build_gap_sequence (seq, s, newd, m,  nlist, rlist, ptab, threads);
				if ( n < k0 ) {
					clear_gap_sequence (seq);
					n = build_gap_sequence (seq, s, *pd, m,  nlist, rlist, ptab, threads);
					break;
				}
				*pd = newd;
			}
			if ( *pd != d ) printf ("%s d to %ld (%.1fs)\n", (d < *pd ? "increased" : "decreased" ), *pd, omp_get_wtime()-begin);
			return n;
		}
		clear_gap_sequence (seq);
		newd = ((((long)(1.001*(*pd)+1))>>6)+1)<<6;
		if ( newd == *pd ) newd += 64;
		*pd = newd;
	}
}

// return the least minimally occupied residue modulo p, or -1 if sequence is admissible mod p
long gap_sequence_best_residue (gap_sequence_t seq, long p)
{
	register int *map, *ap, *ep;
	register unsigned char *gap;
	register long a;
	register int n;

	assert ( p > 510 );
	map = calloc (p, sizeof(*map));  ep = map+p;
	a = seq->s;
	ap = map + mod (seq->s, p);  (*ap)++;
	gap = seq->gaps;
	// process initial sequence of gaps of size < 510
	while ( gap < seq->firstz ) { ap += 2*(*gap++); if ( ap >= ep ) ap -= p;  (*ap)++;}
	// process middle sequence of gaps that may be >= 510
	while ( gap < seq->endz ) { while ( ! *gap ) { ap += 510; if ( ap >= ep ) ap -= p; gap++; } ap += 2*(*gap++); if ( ap >= ep ) ap -= p;  (*ap)++; }
	// process final sequence of gaps of size < 510
	while ( gap < seq->egap ) { ap += 2*(*gap++); if ( ap >= ep ) ap -= p;  (*ap)++; }
	for ( n = map[0], a = 0, ap = map+1 ; ap < ep ; ap++ ) if ( *ap < n ) { n = *ap; a = ap-map; }
	free (map);
	return ( n ? a : -1 );
}

// sets rm[n] to the least occupied residue class mod p_{m+1+n}, or to -1 if admissible at p_{m+1+n}, for n from 0 to checks-1
// returns count of admissible primes in p_{m+1} to p_{m+1+checks}
long gap_sequence_residue_map (long *rm, gap_sequence_t seq, prime_table_t ptab, long m, long checks, long maxm, int threads)
{
	long w = 0;
	
	assert ( m <= maxm );
	if ( checks > maxm-m ) { for ( long i = maxm-m ; i < checks ; i++ ) { rm[i] = -1; w++; } checks = maxm-m; }

	#pragma omp parallel num_threads(threads)
	{
		int tid = omp_get_thread_num ();
		long r, n;
		
		for ( n = tid ; n < checks ; n += threads ) {
			r = gap_sequence_best_residue (seq, prime_table_nth_prime (ptab, m+1+n));
			rm[n] = r;
		}
	}
	for ( long i = 0 ; i < checks ; i++ ) if ( rm[i] == -1 ) w++;
	return w;
}

// return the least minimally occupied residue modulo p, or -1 if sequence is admissible mod p
int gap_sequence_admissible_mod_p (gap_sequence_t seq, long p)
{
	uint64_t *bm;
	register long ap;
	register unsigned char *gap;
	int sts;

	assert ( p > 510 );
	bm = bm_alloc (p);
	
	ap = mod (seq->s, p);  bm_set (bm, ap);
	gap = seq->gaps;
	// process initial sequence of gaps of size < 510
	while ( gap < seq->firstz ) { ap += 2*(*gap++); if ( ap >= p ) ap -= p; bm_set (bm, ap); }
	// process middle sequence of gaps that may be >= 510
	while ( gap < seq->endz ) { while ( ! *gap ) { ap += 510; if ( ap >= p ) ap -= p; gap++; } ap += 2*(*gap++); if ( ap >= p ) ap -= p; bm_set (bm, ap); }
	// process final sequence of gaps of size < 510
	while ( gap < seq->egap ) { ap += 2*(*gap++); if ( ap >= p ) ap -= p; bm_set (bm, ap); }
	
	sts = bm_allset (bm, p);
	bm_free (bm);
	return ! sts;
}

long gap_sequence_best_subsequence (long *ps, long *pt, gap_sequence_t seq, long k0)
{
	register unsigned char *sgap, *tgap;
	register long k, s, t, bests, bestt, d;
	
	s = seq->s;  sgap = seq->gaps;
	t = seq->t;  tgap = seq->egap-1;
	for ( k = seq->k ; k > k0 ; k-- ) { t -= 2*(*tgap--);  while ( ! *tgap ) { t-=510; tgap--; } }
	tgap++;
	bests = s;  bestt = t;  d = t-s;
	while ( tgap < seq->egap ) {
		while ( !*sgap ) { s+=510; sgap++; }
		s += 2*(*sgap++);
		while ( !*tgap ) { t+=510; tgap++; }
		t += 2*(*tgap++);
		if ( t-s < d ) { bests = s; bestt = t; d = t-s; }
	}
	*ps = bests;  *pt = bestt;
	return d;
}

void dump_gap_sequence (gap_sequence_t seq)
{
	char filename[256];
	FILE *fp;
	register unsigned char *gap;
	register long i, n;
	
	sprintf (filename, "sequence_%ld_%ld.txt", seq->k, seq->t-seq->s);
	fp = fopen (filename, "w");
	n = seq->s;
	fprintf (fp, "%ld\n", n);
	gap = seq->gaps;
	for ( i = 1 ; gap < seq->egap ; i++ ) {
		while ( !*gap ) { n += 510; gap++; }
		n += 2*(*gap++);
		fprintf (fp, "%ld\n", n);
	}
	fclose (fp);
	assert ( i == seq->k );
}

// sieves the interval [s,s+d-1] of 1 mod 2 and 0 mod p for primes 3,...,p_n, and optionally at rlist[i] mod p_{nlist[i]}, where nlist is a zero-terminated list of prime indices n > m
// if rlist is null, sieves at 0 mod each p_{nnlist[i]}
// returns a bitmap of d bits with zero bits indicating survivors
uint64_t *survivor_bitmap (long s, long d, long m, long*nlist, long *rlist, prime_table_t ptab, int threads)
{
	uint64_t *b0, *b1;
	register long blen, bcnt, i, j;

	assert ( m >= 6 );
	j = m;  if ( nlist ) for ( i = 0 ; nlist[i] ; i++ ) if ( nlist[i] > j ) j = nlist[i];
	blen = 2*prime_table_nth_prime (ptab, j);			// aim for block size that is big enough so that we are sieving at least 2 entries for each prime
	if ( d/blen < threads ) blen = d/threads;				// but try to make sure we can give work to every thread
	blen += 64L*15015 - (blen%(64L*15015));			// round b up to a multiple of 64*3*5*7*11*13
	bcnt = d/blen + ( (d%blen) ? 1 : 0 );

	b0 = bm_alloc (blen);
	bitmap_schinzel_sieve (b0, blen, s, 6, ptab);

	b1 = _bm_alloc (d);
	#pragma omp parallel num_threads(threads)
	{
		uint64_t *b;
		int tid = omp_get_thread_num ();
		long e, i;
		
		for ( i = tid ; i < bcnt ; i += threads ) {
			b = b1+i*(blen>>6);
			e = i < bcnt-1 ? blen : d-i*blen;
			memcpy (b, b0, (e>>3) +((e&7)?1:0));
			bitmap_extended_sieve (b, e, s+i*blen, 6, m, nlist, rlist, ptab);
		}
	}
	bm_free (b0);
	bm_complement (b1, d);
	return b1;
}

long survivor_bitmap_best_subsequence (long *ps, uint64_t *bm, long n, long k0)
{
	register long i, j, k, bestd, besti;
	
	i = bm_min_set (bm, n);
	for ( k = 1, j = i ; j < n && k < k0 ; j = bm_next_set (bm, j+1, n) ) k++;
	if ( k < k0 ) return 0;
	bestd = j-i;
	besti = i;
	for ( i = bm_next_set (bm, i+1, n), j = bm_next_set (bm, j+1, n) ; j < n ; i = bm_next_set (bm, i+1, n), j = bm_next_set (bm, j+1, n) ) {
		if ( j-i < bestd ) { bestd = j-i; besti = i; }
	}
	*ps = besti;
	return bestd;
}