#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <omp.h>
#include "bitmap.h"
#include "prime.h"
#include "ompadm.h"

/*
	Copyright (c) 2014 Andrew V. Sutherland
	See LICENSE file for license details
*/

#define MAXB		7
#define L			6	// use 2^L bit words in bitmask array of 2^B words, where B varies dynamically from 0 to MAXB

// determines whether the sequence stored in the specific gap table is admissible modulo p
// pB points to an integer between 0 and MAXB that is used to optimize the test as the density mod p changes
int gtab_admissible_mod_p (gap_table_t tab, long p, int *pB)
{
	uint64_t b[1<<MAXB];
	unsigned char *gap;
	register unsigned char *egap;
	register long j, r, x, o;
	register int n, B;
	
	egap = tab->gaps + tab->ti - tab->si;

	B = *pB;
	o = tab->sp - mod(tab->sp, p) - p;		// o is the greatest multiple of p such that o+p <= sp
	for ( j = 0 ;  j < (p>>(L+B)) ; j++ ) {
		memset (b, 0, (1<<B)*sizeof(*b));
		for ( x = o + (j<<(L+B)) ; x <= tab->tp ; x += p ) {
			for ( r = gap_table_next_gap (tab, &gap, x)-x ; r < (1L<<(L+B)) && gap <= egap ; r+=2*(*gap++) )
				b[(r>>L)&((1<<B)-1)] |= (1UL << (r&((1<<L)-1)));
			for ( n = 0; n < (1<<B) ; n++ ) if ( ~b[n] ) break;
			if ( n == (1<<B) ) break;
		}
		if ( x > tab->tp )  { if ( !j && B ) (*pB)--;  if ( j && B < MAXB ) (*pB)++; return 1; }
	}
	if ( B < MAXB ) (*pB)++;
	for ( j <<= L+B ; j < p ; j++ ) {
		for ( x = o + j ; x <= tab->tp ; x += p ) if ( gap_table_next_gap (tab, &gap, x) == x ) break;
		if ( x > tab->tp ) return 1;
	}
	return 0;
}


int main (int argc, char *argv[])
{
	prime_table_t ptab;
	gap_table_t gtab;
	double begin;
	int threads;
	FILE *fp;
	fpos_t pos;
	char buf[256], *ptr;
	long *Plist, Plistlen;
	long i, j, n, k0, d, m, s, t;
	
	if ( argc < 2 || argc == 4 ) { puts ("schinzelverify k0 s d m [threads]\nschinzelverify filename [threads]"); return 0; }
	if ( argc > 4 ) {
		k0 = atol(argv[1]);
		s = atol (argv[2]);
		d = atol (argv[3]);
		m = atol (argv[4]);
		Plist = 0;
		if ( argc > 5 ) threads = atoi(argv[5]); else threads = 0;
	} else {
		fp = fopen (argv[1],"r");
		if ( ! fp ) { printf ("Unable to open file %s\n", argv[1]); return 0; }
		if ( ! fgets (buf, sizeof(buf), fp) ) { printf ("Error reading line 1 of file %s\n", argv[1]); return 0; }
		if ( sscanf (buf, "k0=%ld", &k0) != 1 )  { printf ("Error parsing line 1 of file %s, expected k0=%%ld\n", argv[1]); return 0; }
		if ( ! fgets (buf, sizeof(buf), fp) ) { printf ("Error reading line 2 of file %s\n", argv[1]); return 0; }
		if ( sscanf (buf, "s=%ld", &s) != 1 )  { printf ("Error parsing line 2 of file %s, expected s=%%ld\n", argv[1]); return 0; }
		if ( ! fgets (buf, sizeof(buf), fp) ) { printf ("Error reading line 3 of file %s\n", argv[1]); return 0; }
		if ( sscanf (buf, "d=%ld", &d) != 1 )  { printf ("Error parsing line 3 of file %s, expected d=%%ld\n", argv[1]); return 0; }
		if ( ! fgets (buf, sizeof(buf), fp) ) { printf ("Error reading line 4 of file %s\n", argv[1]); return 0; }
		if ( sscanf (buf, "m=%ld", &m) != 1 && sscanf (buf, "n=%ld", &m) != 1 )  { printf ("Error parsing line 4 of file %s, expected m=%%ld\n", argv[1]); return 0; }
		if ( fgetpos (fp, &pos) ) { printf ("fgetpost failed on file %s\n", argv[1]); return 0; }
		Plistlen = 0;
		while ( fgets(buf, sizeof(buf), fp) ) {
			if ( (ptr = strstr (buf, "-")) ) { *ptr = 0;  i = atol (buf);  j = atol(ptr+1); } else { i = j = atol(buf); }
			Plistlen += j-i+1;
		}
		if ( fsetpos (fp, &pos) ) { printf ("fsetpost failed on file %s\n", argv[1]); return 0; }	
		Plist = malloc ((Plistlen+1)*sizeof(*Plist));
		Plistlen = 0;
		while ( fgets(buf, sizeof(buf), fp) ) {
			if ( (ptr = strstr (buf, "-")) ) { *ptr = 0;  i = atol (buf);  j = atol(ptr+1); } else { i = j = atol(buf); }
			for ( Plist[Plistlen] = i ; Plist[Plistlen] < j ; Plist[Plistlen+1] = Plist[Plistlen]+1, Plistlen++ );
			Plistlen++;
		}
		Plist[Plistlen] = 0;
		fclose (fp);
		if ( argc > 2 ) threads = atoi(argv[2]); else threads = 0;
	}
	
	if ( threads <= 0 || threads > omp_get_max_threads() ) threads = omp_get_max_threads();
	printf ("omp threads = %d\n", threads);
	
	prime_table_init (ptab, 2, prime_pi(k0)+1);	// we need to go one prime past k0
	
	if ( Plist ) {
		printf ("k0=%ld, s=%ld, d=%ld, p_%ld=%ld (and %ld of the next %ld primes)\n", k0, s, d, m, prime_table_nth_prime(ptab, m), Plistlen, Plist[Plistlen-1]-m);
	} else {
		printf ("k0=%ld, s=%ld, d=%ld, p_%ld=%ld\n", k0, s, d, m, prime_table_nth_prime(ptab, m));
	}
	
	begin = omp_get_wtime();
	n = build_gap_table (gtab, s, d+1, m, Plist, 0, ptab, threads);
	printf ("Schinzel sieve  of [%ld,%ld] for p <= p_%ld = %ld yielded %ld survivors (%.1fs)\n", gtab->sp, gtab->tp, m, prime_table_nth_prime (ptab, m), n, omp_get_wtime()-begin);
	if ( n < k0 ) { printf ("Not enough survivors, you need to increase the diameter or decrease the prime bound\n"); return 0; }

	#pragma omp parallel num_threads(threads)
	{
		double start;
		unsigned char *pgap;
		int tid, B;
		register long p;

		tid = omp_get_thread_num();
		start = omp_get_wtime();
		p = prime_table_nth_gap (ptab, &pgap, m);
		for ( int  i = 0 ; p <= k0 && i < tid ; i++ ) p += 2*(*pgap++);
		B = MAXB;
		while ( p <= k0 ) {
			if ( ! gtab_admissible_mod_p (gtab, p, &B) ) { printf ("Inadmissible modulo p_%ld=%ld, terminating.\n", prime_table_pi (ptab, p), p); exit(-1); }
			#pragma omp master
			{
				double end = omp_get_wtime();
				if ( end-start > 3 ) { printf ("verified to p_%ld = %ld (%.1fs)\n", prime_table_pi (ptab, p), p, end-begin); start = end; }
			}
			for ( int i = 0 ; p <= k0 && i < threads ; i++ ) p += 2*(*pgap++);
		}
	}
	printf ("verified to p = %ld (%.1fs)\n", prime_table_nth_prime (ptab, prime_table_pi (ptab, k0)), omp_get_wtime()-begin);
	
	n = gap_table_best_start (gtab, k0);
	s = gap_table_nth_number (gtab, n);
	t = gap_table_nth_number (gtab, n+k0-1);
	gap_table_clear (gtab);
	if ( Plist ) {
		printf ("Verified admissible Schinzel sequence in [%ld,%ld] with k0=%ld using P_%ld=%ld (and %ld of the next %ld primes), obtaining diameter d=%ld\n", s, t, k0, m, prime_table_nth_prime (ptab, m), Plistlen, Plist[Plistlen-1]-m, t-s);
	} else {
		printf ("Verified admissible Schinzel sequence in [%ld,%ld] with k0=%ld using P_%ld=%ld, obtaining diameter d=%ld\n", s, t, k0, m, prime_table_nth_prime (ptab, m), t-s);
	}
	prime_table_clear (ptab);
	prime_cleanup();

	printf ("Total time %.1fs\n", omp_get_wtime()-begin);
}
