#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <omp.h>
#include "prime.h"

/*
	Copyright (c) 2014 Andrew V. Sutherland
	See LICENSE file for license details
*/

#define MAXB		7
#define L			6	// use 2^L bit words in bitmask array of 2^B words, where B varies dynamically from 0 to MAXB

// determines whether the sequence of primes p_{m+1}, ..., p_{m+k0} is admissible modulo p
// pB points to an integer between 0 and MAXB that is used to optimize the test as the density mod p changes
// requires tab to start at p_{m+1} (this is verified)
int zhang_admissible_mod_p (prime_table_t tab, long k0, long m, long p, int *pB)
{
	uint64_t b[1<<MAXB];
	unsigned char *gap, *egap;
	register long j, r, x, lastp;
	register int i, B;
	
	lastp = prime_table_nth_prime (tab, m+k0);
	assert ( tab->si == m+1 && lastp );
	if ( p < tab->sp || p > lastp ) return 1;
	egap = tab->gaps + k0 - 1;
	B = *pB;
	for ( j = 0 ;  j < (p>>(L+B)) ; j++ ) {
		memset (b, 0, (1<<B)*sizeof(*b));
		for ( x = (j<<(L+B)) ; x <= lastp ; x += p ) {
			for ( r = prime_table_next_gap (tab, &gap, x)-x ; r < (1L<<(L+B)) && gap <= egap ; r+=2*(*gap++) )
				b[(r>>L)&((1<<B)-1)] |= (1UL << (r&((1<<L)-1)));
			for ( i = 0; i < (1<<B) ; i++ ) if ( ~b[i] ) break;
			if ( i == (1<<B) ) break;
		}
		if ( x > lastp ) { if ( !j && B ) (*pB)--;  if ( j && B < MAXB ) (*pB)++; return 1; }
	}
	if ( B < MAXB ) (*pB)++;
	for ( j <<= L+B ; j < p ; j++ ) {
		for ( x = j ; x <= lastp ; x += p ) if ( prime_table_next_gap (tab, &gap, x) == x ) break;
		if ( x > lastp ) return 1;
	}
	return 0;
}


int main (int argc, char *argv[])
{
	prime_table_t tab;
	double begin;
	int threads;
	register long k0, m;

	if ( argc < 3 ) { puts ("zhangverify k0 m [threads]\n    verifies that p_{m+1},...,p_{m+k_0} is admissible."); return 0; }
	k0 = atol(argv[1]);
	assert ( k0 > 2 );
	m = atol(argv[2]);
	assert ( m >= 1 );
	if ( argc > 3 ) threads = atoi(argv[3]); else threads = omp_get_max_threads();
	if ( threads <= 0 || threads > omp_get_max_threads() ) threads = omp_get_max_threads();
	printf ("omp threads = %d\n", threads);
	
	begin = omp_get_wtime();
	prime_table_init (tab, m+1, m+k0);
	printf ("Built prime table from p_%ld=%ld to p_%ld=%ld (%.1fs)\n", tab->si, tab->sp, tab->ti, tab->tp, omp_get_wtime()-begin);

	// Note that we only need to check admissible for primes in [p_{m+1},k0] since no element of p_{m+1},...,p_{m+k0} is zero mod p for p < p_{m+1} and admissibility mod p > k0 is automatic.
	#pragma omp parallel num_threads(threads)
	{
		double start;
		unsigned char *pgap;
		int tid, B;
		register long p;

		tid = omp_get_thread_num();
		start = omp_get_wtime();
		p = tab->sp;  pgap = tab->gaps;
		for ( int  i = 0 ; p <= k0 && i < tid ; i++ ) p += 2*(*pgap++);
		B = MAXB;
		while ( p <= k0 ) {
			if ( ! zhang_admissible_mod_p (tab, k0, m, p, &B) ) { printf ("Inadmissible modulo p_%ld=%ld, terminating.\n", prime_table_pi (tab, p), p); exit(-1); }
			#pragma omp master
			{
				double end = omp_get_wtime();
				if ( end-start > 3 ) { printf ("verified to p = %ld (%.1fs)\n", p, end-begin); start = end; }
			}
			for ( int i = 0 ; p <= k0 && i < threads ; i++ ) p += 2*(*pgap++);
		}
	}
	printf ("verified to p = %ld (%.1fs)\n", prime_table_nth_prime (tab, prime_table_pi (tab, k0)), omp_get_wtime()-begin);
	printf ("Zhang sequence p_{m+1},...,p_{m+k0} for k0=%ld and m=%ld with diameter %ld is admissible.\n", k0, m, tab->tp-tab->sp);
	prime_table_clear (tab);
	prime_cleanup();
	return 0;
}
